"""CLI interface."""
import itertools
import platform
import shlex
import subprocess
import sys
from pathlib import Path
from typing import List, Optional

import typer
from click.exceptions import UsageError
from packaging.requirements import Requirement
from packaging.version import Version
from rich.console import Console

from idae.pep723 import read
from idae.resolver import get_python_or_exit
from idae.venv import Python, clean_venvs, get_venv

if sys.version_info < (3, 9):  # pragma: no cover
    from typing_extensions import Annotated
else:
    from typing import Annotated
cli = typer.Typer()

console = Console(stderr=True)


@cli.command(context_settings={"ignore_unknown_options": True})
def run(  # noqa: PLR0913
    script: Annotated[
        Optional[Path],  # noqa: FA100  # Typer can't handle unions
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
            help="The path of the script to run (modules only)",
        ),
    ] = None,
    args: Annotated[
        Optional[List[str]],  # noqa: FA100
        typer.Argument(
            help="Arguments to pass to the script",
        ),
    ] = None,
    python_flags: Annotated[
        Optional[List[str]],  # noqa: FA100
        typer.Option(help="Extra flags to pass to Python"),
    ] = None,
    ignore_version: Annotated[
        bool,
        typer.Option(
            "--ignore-version",
            "-i",
            help="Ignore Python version requirements specified in the script",
        ),
    ] = False,
    clean: Annotated[
        bool,
        typer.Option(
            help="Clean the virtual environment caches",
        ),
    ] = False,
    force_version: Annotated[
        Optional[str],  # noqa: FA100
        typer.Option(
            "--force-version",
            "-f",
            help="Force idae to use a specific Python version",
        ),
    ] = None,
) -> None:
    """Automatically install necessary dependencies to run a Python script.

    --clean can be used without 'SCRIPT'
    """
    if clean:
        clean_venvs()
    if script is None:
        if clean:
            raise typer.Exit(code=0)
        msg = "Missing argument 'SCRIPT'."
        raise UsageError(msg)
    # Get script dependencies
    pyproject = read(str(script.read_text()))
    script_deps = []
    python = Python(
        version=Version(
            f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        ),
        executable=sys.executable,
    )
    if force_version is not None:
        python = get_python_or_exit(force_version, console)
    if pyproject is not None and "run" in pyproject:
        script_deps = (
            []
            if "dependencies" not in pyproject["run"]
            else list(map(Requirement, pyproject["run"]["dependencies"]))
        )

        if (
            not ignore_version
            and force_version is None
            and "requires-python" in pyproject["run"]
        ):
            python = get_python_or_exit(pyproject["run"]["requires-python"], console)

    venv_path = get_venv(script_deps, python)

    extra_flags = list(
        itertools.chain.from_iterable(
            map(shlex.split, python_flags or []),
        ),
    )
    extra_args = list(
        itertools.chain.from_iterable(
            map(shlex.split, args or []),
        ),
    )
    # Run the script inside the venv
    raise typer.Exit(
        code=subprocess.run(
            [  # noqa: S603  # idae is inherently "insecure"
                # If the Python binary doesn't exist, the venv is broken
                str(
                    venv_path
                    / ("Scripts" if platform.system() == "Windows" else "bin")
                    / "python",
                ),
                *extra_flags,
                str(script),
                *extra_args,
            ],
            check=False,
        ).returncode,
    )



"""Dependency utils lifted from Hatch."""
from __future__ import annotations

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from packaging.requirements import Requirement


def _normalize_project_name(project_name: str) -> str:
    # https://peps.python.org/pep-0503/#normalized-names
    return re.sub(r"[-_.]+", "-", project_name).lower()


def _get_normalized_dependency(requirement: Requirement) -> str:
    # Changes to this function affect reproducibility between versions
    from packaging.specifiers import SpecifierSet

    requirement.name = _normalize_project_name(requirement.name)

    if requirement.specifier:
        requirement.specifier = SpecifierSet(str(requirement.specifier).lower())

    if requirement.extras:
        requirement.extras = {
            _normalize_project_name(extra) for extra in requirement.extras
        }

    # All TOML writers use double quotes,
    # so allow direct writing or copy/pasting to avoid escaping
    return str(requirement).replace('"', "'")


def hash_dependencies(requirements: list[Requirement]) -> str:
    """Create a hash of the dependencies."""
    from hashlib import sha256

    data = "".join(
        sorted(
            # Internal spacing is ignored by PEP 440
            normalized_dependency.replace(" ", "")
            for normalized_dependency in {
                _get_normalized_dependency(req) for req in requirements
            }
        ),
    ).encode("utf-8")

    return sha256(data).hexdigest()



"""Resolve Python versions."""

from __future__ import annotations

from typing import TYPE_CHECKING

import findpython  # type: ignore[import-untyped]
import typer
from packaging.specifiers import InvalidSpecifier, SpecifierSet

if TYPE_CHECKING:  # pragma: no cover
    from rich.console import Console


def get_python_or_exit(version: str, console: Console) -> findpython.PythonVersion:
    """Return a PythonVersion or raise Exit."""
    try:
        output = get_python(version)
    except InvalidSpecifier as err:
        console.print(f"[red]error: Python version {version} could not be parsed[/red]")
        raise typer.Exit(code=1) from err
    if output is None:
        console.print(f"[red]error: Python version {version} not found[/red]")
        raise typer.Exit(code=1)
    return output


def get_python(version: str) -> findpython.PythonVersion | None:
    """Resolve the version string and return a valid Python."""
    # Order from latest version to earliest
    pythons = {python.version: python for python in findpython.find_all()}
    try:
        float(version)
    except ValueError:
        pass
    else:
        version = f"~={float(version)}"
    target = SpecifierSet(version)
    for python_version, python in pythons.items():
        if python_version in target:
            return python
    return None



"""Custom parser following the specification from PEP 723."""
from __future__ import annotations

import re
from typing import Any

try:
    import tomllib  # type: ignore[import, unused-ignore]
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore[import, unused-ignore, no-redef]
COMMENT_BLOCK_REGEX = re.compile(
    r"^\s*#\s*///\s*(?P<type>[a-zA-Z0-9-]+)$\s*(?P<content>(?:^#.*$\s)+)^\s*#\s*///$",
    flags=re.MULTILINE,
)
COMMENT_REGEX = re.compile(r"^\s*#\s*(.+)$", flags=re.MULTILINE)


def read_comment_block(script: str, name: str) -> str | None:
    """Find and returns the contents of the metadata comment block."""
    matches = list(
        filter(
            lambda m: m.group("type") == name,
            re.finditer(COMMENT_BLOCK_REGEX, script),
        ),
    )
    if len(matches) > 1:
        msg = f"Multiple {name} blocks found"
        raise ValueError(msg)
    if len(matches) == 1:
        return "\n".join(COMMENT_REGEX.findall(matches[0].group("content")))
    return None


def read(script: str) -> Any:  # noqa: ANN401
    """Find and returns the value of the pyproject comment block."""
    comment_block = read_comment_block(script, "pyproject")

    return tomllib.loads(comment_block) if comment_block is not None else comment_block



"""Utils for venv creation."""
from __future__ import annotations

import platform
import shutil
import subprocess
from dataclasses import dataclass
from typing import TYPE_CHECKING

import platformdirs

if TYPE_CHECKING:  # pragma: no cover
    from os import PathLike
    from pathlib import Path

    from packaging.requirements import Requirement
    from packaging.version import Version

from .dependencies import hash_dependencies

CACHE_DIR = platformdirs.user_cache_path("idae")


@dataclass
class Python:
    """Object representing a Python executable."""

    version: Version
    executable: str | PathLike[str]


def get_venv(requirements: list[Requirement], python: Python) -> Path:
    """Create or fetch a cached venv."""
    dep_hash = hash_dependencies(requirements)
    venv_path = CACHE_DIR / f"{python.version.major}.{python.version.minor}" / dep_hash
    if venv_path.is_dir():
        return venv_path
    # This automatically includes pip
    subprocess.run(
        [python.executable, "-m", "venv", venv_path],  # noqa: S603
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True,
    )
    # Install dependencies into the venv (if any)
    if requirements:
        subprocess.run(
            [  # noqa: S603
                (
                    venv_path
                    / ("Scripts" if platform.system() == "Windows" else "bin")
                    / "pip"
                ).resolve(),
                "install",
                *map(str, requirements),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
        )
    # The above works according to the Python docs:
    # > You don't specifically need to activate a virtual environment,
    # > as you can just specify the full path to that environment`s Python interpreter
    # > when invoking Python. Furthermore, all scripts installed in the environment
    # > should be runnable without activating it.
    # - https://docs.python.org/3/library/venv.html#how-venvs-work
    return venv_path


def clean_venvs() -> None:
    """CLI command to delete the cache."""
    # Ignore errors like the directory not existing
    shutil.rmtree(CACHE_DIR, ignore_errors=True)
